# /travisbumpversion/main.py
#
# Entry point for travis-bump-version.
#
# See /LICENCE.md for Copyright information
"""Entry point for travis-bump-version.."""
import argparse

import atexit

import os

import subprocess

import sys

try:
    from subprocess import DEVNULL  # suppress(ungrouped-imports)
except ImportError:
    DEVNULL = open(os.devnull, "wb")
    atexit.register(DEVNULL.close)


def error(message):
    """Print message to stderr."""
    sys.stderr.write("{}\n".format(message))


def _last_commit_message_body():
    """Read body of last commit message from git."""
    return subprocess.check_output(["git",  # pragma: no cover
                                    "log",
                                    "--format=%b",
                                    "-n",
                                    "1"]).decode("utf-8").strip()


def _get_level():
    """Read the last commit message and get the version bump level."""
    level = "patch"
    last_msg = _last_commit_message_body()

    if len(last_msg):
        code = [w for w in last_msg.splitlines()[-1].split(" ") if len(w)]

        if len(code) > 1 and code[0] == "bumpversion:":
            valid = ("major", "minor", "patch")
            if code[1] not in valid:
                error("""bumpversion: Invalid code {} given, expected """
                      """major, minor or patch. """
                      """Assuming patch""".format(code[1]))
            else:
                level = code[1]

    return level


def _last_tag_from_git(repo=""):
    """Read known tag from git."""
    subprocess.check_call(["git", "checkout", "master"])  # pragma: no cover
    subprocess.check_call(["git",  # pragma: no cover
                           "fetch",
                           "--tags",
                           "git://github.com/{}".format(repo)])
    return subprocess.check_output(["git",  # pragma: no cover
                                    "describe",
                                    "--tags"]).decode("utf-8").strip()


def _call_bumpversion(level, files):
    """Call bumpversion to create commit and bump version."""
    subprocess.check_call(["git",  # pragma: no cover
                           "config",
                           "user.name",
                           "Travis-CI"])
    subprocess.check_call(["git",  # pragma: no cover
                           "config",
                           "user.email",
                           "travis@travis-ci.org"])
    subprocess.check_call([  # pragma: no cover
        "bumpversion",
        level
    ] + list(files) + [
        "--commit",
        "--message",
        "Bump version: {current_version} -> {new_version}\n\n[ci skip]",
        "--tag",
        "--verbose"
    ])


# pragma: no cover
def _push_commit_and_tags(api_token, repo):
    """Push most recent commit and tags to repo, using api_token."""
    url = "https://{0}:x-oauth-basic@github.com/{1}.git".format(api_token,
                                                                repo)
    process = subprocess.Popen(["git",
                                "push",
                                url,
                                "master",
                                "--tags"],
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)

    stdout, stderr = process.communicate()

    if process.returncode != 0:
        raise RuntimeError("""Failed with {returncode}, """
                           """{stdout} {stderr}"""
                           """""".format(returncode=process.returncode,
                                         stdout=stdout,
                                         stderr=stderr))


def version_bump(api_token=None,
                 repo=None,
                 files=None):
    """Check if possible to bump version and bump version."""
    if not files or not len(files):
        error("""Need to specify some files""")
        return 1

    last_tag = _last_tag_from_git(repo)

    if not last_tag.startswith("v"):
        error("""Release a tag that matches vx.x.x to get automatic """
              """version bumps""")
        return 1

    _call_bumpversion(_get_level(), files)

    try:
        _push_commit_and_tags(api_token, repo)
    except RuntimeError as exception:
        error("""bumpversion: Failed to push commit, assuming that another """
              """job was able to push it first. The reason given was:\n"""
              """{}""".format(str(exception).replace(api_token, "[redacted]")))

    return 0


def main(arguments=None):  # suppress(unused-function)
    """Run travis-bump-version, pass files to version_bump."""
    parser = argparse.ArgumentParser(description="""Automatically bump """
                                                 """version on successful """
                                                 """Travis-CI builds.""")
    parser.add_argument("--api-token",
                        required=True,
                        type=str,
                        help="""GitHub API token. Always make sure this """
                             """variable is encrypted first.""")
    parser.add_argument("--repo",
                        required=True,
                        type=str,
                        help="""Name of repo on GitHub, eg user/repo.""")
    parser.add_argument("files",
                        nargs="*",
                        metavar=("FILE"),
                        help="""bump version on FILE""")
    result = parser.parse_args(arguments or sys.argv[1:])

    sys.exit(version_bump(**vars(result)))



# /test/test_acceptance.py
#
# Acceptance tests for travis-bump-version.
#
# See /LICENCE.md for Copyright information
"""Acceptance tests for travis-bump-version."""

import sys

from mock import Mock, call

from nose_parameterized import parameterized

from six import StringIO

from testtools import ExpectedException, TestCase
from testtools.matchers import Contains, MatchesAll, Not

import travisbumpversion.main


def run(*args, **kwargs):
    """Run, converting kwargs to switches and passing args."""
    arguments = list(args)
    for key, value in kwargs.items():
        arguments.append("--" + key.replace("_", "-"))
        if not isinstance(value, list):
            value = [value]

        arguments.extend(value)

    try:
        travisbumpversion.main.main(arguments)
    finally:
        sys.stderr.seek(0)
        sys.stdout.seek(0)


def _format_with_args(func, num, params):
    """Format docstring using provided arguments."""
    del num

    return func.__doc__.format(*params[0])


class TestAcceptance(TestCase):
    """Acceptance tests for travis-bump-version."""

    def __init__(self, *args, **kwargs):
        """Initialize instance variables."""
        super(TestAcceptance, self).__init__(*args, **kwargs)
        self.last_commit_message_body = Mock()
        self.last_tag_from_git = Mock()
        self.call_bumpversion = Mock()
        self.push_commit_and_tags = Mock()

    def setUp(self):
        """Set up acceptance tests, patch out functions calling subprocess."""
        super(TestAcceptance, self).setUp()
        self.patch(travisbumpversion.main,
                   "_last_commit_message_body",
                   self.last_commit_message_body)
        self.patch(travisbumpversion.main,
                   "_last_tag_from_git",
                   self.last_tag_from_git)
        self.patch(travisbumpversion.main,
                   "_call_bumpversion",
                   self.call_bumpversion)
        self.patch(travisbumpversion.main,
                   "_push_commit_and_tags",
                   self.push_commit_and_tags)
        self.patch(sys, "exit", Mock())
        self.patch(sys, "stdout", StringIO())
        self.patch(sys, "stderr", StringIO())

    def test_bail_when_not_passing_api_token(self):
        """Raise when failing to pass --api-token."""
        with ExpectedException(RuntimeError):
            sys.exit.side_effect = RuntimeError("")
            run("file", repo="user/repo")

        output = sys.stderr.read()
        self.assertThat(output, MatchesAll(Contains("""--api-token"""),
                                           Contains("""required""")))
        self.assertEqual(call(2), sys.exit.call_args_list[0])

    def test_bail_when_not_passing_repo(self):
        """Raise when failing to pass --repo."""
        with ExpectedException(RuntimeError):
            sys.exit.side_effect = RuntimeError("")
            run("file", api_token="token")

        output = sys.stderr.read()
        self.assertThat(output, MatchesAll(Contains("""--repo"""),
                                           Contains("""required""")))
        self.assertEqual(call(2), sys.exit.call_args_list[0])

    def test_bail_when_not_passing_files(self):
        """Raise when failing to pass list of files."""
        run(api_token="token", repo="user/repo")
        self.assertThat(sys.stderr.read(),
                        Contains("""Need to specify some files"""))
        self.assertEqual(call(1), sys.exit.call_args_list[0])

    def test_bail_when_not_last_tag_doesnt_start_with_v(self):
        """Raise when last tag doesn't start with v."""
        self.last_tag_from_git.return_value = "x0.0.1"
        run("file", api_token="token", repo="user/repo")
        self.assertThat(sys.stderr.read(),
                        Contains("""Release a tag"""))
        self.assertEqual(call(1), sys.exit.call_args_list[0])

    def test_use_patch_by_default_for_empty_body(self):
        """Use level patch by default for empty commit message body."""
        self.last_tag_from_git.return_value = "v0.0.1"
        self.last_commit_message_body.return_value = ""
        run("file", api_token="token", repo="user/repo")
        self.call_bumpversion.assert_called_with("patch", ["file"])
        self.assertEqual(call(0), sys.exit.call_args_list[0])

    def test_use_patch_by_default_for_body_with_no_code(self):
        """Use level patch by default commit message body with no code."""
        self.last_tag_from_git.return_value = "v0.0.1"
        self.last_commit_message_body.return_value = "Title\n\nBody"
        run("file", api_token="token", repo="user/repo")
        self.call_bumpversion.assert_called_with("patch", ["file"])
        self.assertEqual(call(0), sys.exit.call_args_list[0])

    def test_use_patch_level_for_malformed_bumpversion_code(self):
        """Use level patch when bumpversion code is malformed."""
        self.last_tag_from_git.return_value = "v0.0.1"
        self.last_commit_message_body.return_value = "T\n\nbumpversion: mjr"
        run("file", api_token="token", repo="user/repo")
        self.call_bumpversion.assert_called_with("patch", ["file"])
        self.assertEqual(call(0), sys.exit.call_args_list[0])

    def test_complain_for_malformed_bumpversion_code(self):
        """Complain when bumpversion code is malformed."""
        self.last_tag_from_git.return_value = "v0.0.1"
        self.last_commit_message_body.return_value = "T\n\nbumpversion: mjr"
        run("file", api_token="token", repo="user/repo")
        self.assertThat(sys.stderr.read(), Contains("""Invalid code mjr"""))

    @parameterized.expand(("major", "minor", "patch"),
                          testcase_func_doc=_format_with_args)
    def test_use_valid_bumpversion_code(self, lvl):
        """Use level {0} when bumpversion code is set to {0}."""
        self.last_tag_from_git.return_value = "v0.0.1"
        self.last_commit_message_body.return_value = "T\n\nbumpversion: " + lvl
        run("file", api_token="token", repo="user/repo")
        self.call_bumpversion.assert_called_with(lvl, ["file"])
        self.assertEqual(call(0), sys.exit.call_args_list[0])

    def test_push_commit_and_tags_success(self):
        """Push commit and tags with provided token and repo."""
        self.last_tag_from_git.return_value = "v0.0.1"
        self.last_commit_message_body.return_value = ""
        run("file", api_token="token", repo="user/repo")
        self.push_commit_and_tags.assert_called_with("token", "user/repo")
        self.assertEqual(call(0), sys.exit.call_args_list[0])

    def test_push_commit_and_tags_complain_on_failure(self):
        """Push commit and tags with provided token and repo."""
        self.last_tag_from_git.return_value = "v0.0.1"
        self.last_commit_message_body.return_value = ""
        error = RuntimeError("""Failed""")
        self.push_commit_and_tags.side_effect = error
        run("file", api_token="token", repo="user/repo")
        self.assertThat(sys.stderr.read(), Contains("""Failed to push"""))
        self.assertEqual(call(0), sys.exit.call_args_list[0])

    def test_push_commit_fail_contains_git_reasons(self):
        """Failed push reports why git push failed."""
        self.last_tag_from_git.return_value = "v0.0.1"
        self.last_commit_message_body.return_value = ""
        error = RuntimeError("""Git reasons""")
        self.push_commit_and_tags.side_effect = error
        run("file", api_token="token", repo="user/repo")
        self.assertThat(sys.stderr.read(), Contains("""Git reasons"""))
        self.assertEqual(call(0), sys.exit.call_args_list[0])

    def test_push_failure_dont_disclose_api_key(self):
        """Don't disclose API keys on push failure."""
        self.last_tag_from_git.return_value = "v0.0.1"
        self.last_commit_message_body.return_value = ""
        error = RuntimeError("""Failed token""")
        self.push_commit_and_tags.side_effect = error
        run("file", api_token="token", repo="user/repo")
        stderr_output = sys.stderr.read()
        self.assertThat(stderr_output, Not(Contains("token")))
        self.assertThat(stderr_output, Contains("[redacted]"))
        self.assertEqual(call(0), sys.exit.call_args_list[0])



# /setup.py
#
# Installation and setup script for travis-bump-version
#
# See /LICENCE.md for Copyright information
"""Installation and setup script for travis-bump-version."""

from setuptools import find_packages
from setuptools import setup

setup(name="travis-bump-version",
      version="0.1.11",
      description="Bump version files on travis builds",
      long_description_markdown_filename="README.md",
      author="Sam Spilsbury",
      author_email="smspillaz@gmail.com",
      url="http://github.com/polysquare/travis-runner",
      classifiers=["Development Status :: 3 - Alpha",
                   "Programming Language :: Python :: 2",
                   "Programming Language :: Python :: 2.7",
                   "Programming Language :: Python :: 3",
                   "Programming Language :: Python :: 3.1",
                   "Programming Language :: Python :: 3.2",
                   "Programming Language :: Python :: 3.3",
                   "Programming Language :: Python :: 3.4",
                   "Intended Audience :: Developers",
                   "Topic :: Software Development :: Build Tools",
                   "License :: OSI Approved :: MIT License"],
      license="MIT",
      keywords="development linters",
      packages=find_packages(exclude=["tests"]),
      install_requires=["pyaml", "bumpversion"],
      extras_require={
          "green": ["nose-parameterized",
                    "testtools",
                    "six",
                    "setuptools-green"],
          "polysquarelint": ["polysquare-setuptools-lint>=0.0.19"],
          "upload": ["setuptools-markdown"]
      },
      entry_points={
          "console_scripts": [
              "travis-bump-version=travisbumpversion.main:main"
          ]
      },
      test_suite="nose.collector",
      zip_safe=True,
      include_package_data=True)
